import os
from io import BytesIO

from PIL import Image
from dotenv import load_dotenv
from minio import Minio

import streetview
from routeparser import get_points_from_coordinates


class Config:
    MINIO_ENDPOINT: str
    MINIO_ACCESS_KEY: str
    MINIO_SECRET_KEY: str


def load_config(env_path: str = ".env") -> Config:
    if os.path.exists(env_path):
        load_dotenv(env_path)

    cfg = Config()

    cfg.MINIO_ENDPOINT = os.getenv("MINIO_ENDPOINT")
    cfg.MINIO_ACCESS_KEY = os.getenv("MINIO_ACCESS_KEY")
    cfg.MINIO_SECRET_KEY = os.getenv("MINIO_SECRET_KEY")

    return cfg


def download_s3(
        start_lat: float,
        start_lon: float,
        end_lat: float,
        end_lon: float,
        minio_client: Minio,
        bucket_name: str,
        folder_name: str
):
    points = get_points_from_coordinates(start_lat, start_lon, end_lat, end_lon)
    panoids_buffer = []
    print(len(points))

    for point in points:
        panoids = streetview.panoids(lat=point[0], lon=point[1], closest=True)

        if len(panoids) == 0 or panoids[-1]["panoid"] in panoids_buffer:
            continue

        panorama = streetview.download_panorama_v3(panoids[-1]['panoid'], zoom=2, disp=False)
        panoids_buffer.append(panoids[-1]['panoid'])
        image = Image.fromarray(panorama)
        buf = BytesIO()
        image.save(buf, format='PNG')
        buf.seek(0)

        minio_client.put_object(
            bucket_name,
            f"{folder_name}/{panoids[-1]['lat']}_{panoids[-1]['lon']}.png",
            buf,
            buf.getbuffer().nbytes,
            content_type="image/png"
        )

    return len(points)


if __name__ == "__main__":
    config = load_config()

    lat, lon = 55.61727443131236, 37.61265372859094
    lat2, lon2 = 55.618037091058774, 37.61304548448152

    minio_client = Minio(
        endpoint=config.MINIO_ENDPOINT,
        access_key=config.MINIO_ACCESS_KEY,
        secret_key=config.MINIO_SECRET_KEY
    )

    download_s3(lat, lon, lat2, lon2, minio_client, "test", "industrial_images")
